__version__ = "0.0.6"

from .core import (
    broadcast_batch,
    to_hom,
    homogenize,
    compose_rt,
    to_44,
    to_34,
    look_at_np,
    to_np,
    to_numpy,
    to_torch,
    to_8b,
    to_float,
    rotx,
    roty,
    rotz,
    map_range,
    create_random_cameras_on_unit_sphere,
    opengl_c2w_to_opencv_c2w,
    opencv_c2w_to_opengl_c2w,
    opengl_project_from_opencv_intrinsics,
)

from .geometry_basic import (
    duplicate_faces,
    remove_duplicate_faces,
    get_aspect_ratio,
    normalize_vertices,
    clean_infinite_vertices,
    calc_edges,
    calc_face_normals,
    calc_vertex_normals,
    calc_edge_length,
    get_center_of_attention,
    scale_poses,
    find_princple_componenets,
    rotmat2qvec,
    qvec2rotmat,
    qslerp
)

from .geometry_advanced import (
    distribute_field,
    distribute_scalar_field,
    distribute_vector_field,
    qem
)

from .gsoup_io import (
    write_to_json,
    save_image,
    save_images,
    save_animation,
    save_mesh,
    save_meshes,
    save_obj,
    load_image,
    load_images,
    load_obj,
    load_mesh,
)

from .image import (
    alpha_compose,
    draw_text_on_image,
    draw_gizmo_on_image,
    merge_figures_with_line,
    generate_voronoi_diagram,
    interpolate_single_channel,
    interpolate_multi_channel,
    image_grid,
    resize_images_naive,
    change_brightness,
    change_contrast
)

from .video import (
    get_video_info,
    get_frame_timestamps,
    load_video,
    save_video,
    reverse_video,
    compress_video,
    video_to_images,
    slice_from_video,
    VideoReader,
    FPS
)

from .procam import (
    warp_image,
    generate_gray_code,
    pix2pix_correspondence,
    calibrate_procam
)

from .sphere_trace import (
    generate_rays,
    render
)

from . import structures


import time
import collections
import numpy as np
from pathlib import Path
import ffmpeg
from .image import resize_images_naive

class FPS:
    """
    calculates current fps and returns it, see https://stackoverflow.com/a/54539292
    example usage:
        fps = FPS()
        while True:
            # do something
            print(fps())
    """
    def __init__(self,avarageof=50):
        self.frametimestamps = collections.deque(maxlen=avarageof)
    def __call__(self):
        self.frametimestamps.append(time.time())
        if(len(self.frametimestamps) > 1):
            return len(self.frametimestamps)/(self.frametimestamps[-1]-self.frametimestamps[0])
        else:
            return 0.0

def get_video_info(video_path):
    """
    returns basic video info
    :param video_path: path to video
    :return: height, width, fps
    """
    video_path = Path(video_path)
    probe = ffmpeg.probe(str(video_path))
    video_stream = next((stream for stream in probe['streams'] if stream['codec_type'] == 'video'), None)
    width = int(video_stream['width'])
    height = int(video_stream['height'])
    fps = float(video_stream['r_frame_rate'].split('/')[0]) / float(video_stream['r_frame_rate'].split('/')[1])
    frame_count = int(video_stream['nb_frames'])
    # codec_name = video_stream['codec_name']
    # format_name = probe['format']['format_name']
    return height, width, fps, frame_count

def get_frame_timestamps(video_path):
    """
    returns numpy array of frame timestamps in seconds for a video
    :param video_path: path to video
    :return: numpy array of frame timestamps in seconds
    """
    video_path = Path(video_path)
    probe = ffmpeg.probe(str(video_path), show_frames="-show_frames")
    video_frames = [frame for frame in probe['frames'] if frame['media_type'] == 'video']
    frame_times = np.array([float(frame["pts_time"]) for frame in video_frames])
    return frame_times


def load_video(video_path):
    """
    loads a video from disk into a numpy tensor (uint8, channels last, RGB)
    :param video_path: path to video
    :return: (n x h x w x 3) tensor
    """
    h, w, _, _ = get_video_info(video_path)
    out, _ = (
        ffmpeg
        .input(str(video_path))
        .output('pipe:', format='rawvideo', pix_fmt='rgb24')
        .run(capture_stdout=True, quiet=True)
    )
    video = np.frombuffer(out, np.uint8).reshape([-1, h, w, 3])
    return video

def save_video(frames, output_path, fps, lossy=False):
    """
    saves a video from a t x h x w x 3 numpy tensor
    :param frames: (t x h x w x 3) tensor
    :param output_path: path to save video to
    :param fps: frames per second of output video
    :param lossy: if True, use lossy compression (default: False, but then only .avi is supported)
    """
    output_path = Path(output_path)
    output_path.parent.mkdir(parents=True, exist_ok=True)
    if lossy:
        (
        ffmpeg
        .input('pipe:', format='rawvideo', pix_fmt='rgb24', s='{}x{}'.format(frames.shape[2], frames.shape[1]), r=fps)
        .output(str(output_path), pix_fmt='yuv420p')
        .overwrite_output()
        .run(input=frames.tobytes(), quiet=True)
        )
    else:
        if output_path.suffix == ".avi":
            pix_fmt = "bgr24"
        else:
            # todo: figure out lossless pixel formats for other containers
            raise ValueError("Lossless video only supported for .avi container")
        (
            ffmpeg
            .input('pipe:', format='rawvideo', pix_fmt='rgb24', s='{}x{}'.format(frames.shape[2], frames.shape[1]), r=fps)
            .output(str(output_path), vcodec='rawvideo', pix_fmt=pix_fmt)
            .overwrite_output()
            .run(input=frames.tobytes(), quiet=True)
        )

def reverse_video(video_path, output_path=None):
    """
    reverses a video and possibly saves it to disk
    :param video_path: path to video
    :param output_path: path to save video to
    :return: (n x h x w x 3) tensor of reversed video
    """
    h, w, _, _ = get_video_info(video_path)
    out, _ = (
        ffmpeg
        .input(str(video_path))
        .filter('reverse')
        .output('pipe:', format='rawvideo', pix_fmt='rgb24')
        .run(capture_stdout=True, quiet=True)
    )
    video = np.frombuffer(out, np.uint8).reshape([-1, h, w, 3])
    if output_path is not None:
        save_video(video, output_path)
    return video

def compress_video(src, dst):
    """
    compresses a video using ffmpeg
    :param src: path to video
    :param dst: path to save compressed video to
    """
    dst = Path(dst)
    dst.parent.mkdir(parents=True, exist_ok=True)
    (
        ffmpeg
        .input(str(src))
        .output(str(dst), vcodec='libx264', crf=23, preset='slow')
        .overwrite_output()
        .run(quiet=True)
    )

def video_to_images(src, dst):
    """
    creates a folder of images from a video
    """
    dst = Path(dst)
    dst.mkdir(parents=True, exist_ok=True)
    (
        ffmpeg
        .input(str(src))
        .output(str(dst / '%d.png'), vcodec='png', format='image2')
        .overwrite_output()
        .run(quiet=True)
    )

def slice_from_video(src, every_n_frames=2, start_frame=0, end_frame=None):
    """
    slices a video into frames
    :param src: path to video
    :param every_n_frames: stride between selected frames
    :param start_frame: first frame to take
    :param end_frame: last frame to take
    :return: (n x h x w x 3) tensor of sliced video
    """
    h, w, _, fc = get_video_info(src)
    if end_frame is None:
        end_frame = fc + 1
    out, _ = (
        ffmpeg
        .input(str(src))
        .filter('select', 'between(n, {}, {})*not(mod(n,{}))'.format(start_frame, end_frame, every_n_frames))
        .output('pipe:', format='rawvideo', pix_fmt='rgb24', fps_mode='passthrough')
        .run(capture_stdout=True, quiet=True)
    )
    video = np.frombuffer(out, np.uint8).reshape([-1, h, w, 3])
    return video

def get_frame_from_video(src, frame_index):
    """
    gets a single frame from a video
    :param src: path to video
    :param frame_num: frame number to get
    :return: (h x w x 3) tensor of frame
    """
    h, w, _, _ = get_video_info(src)
    out, _ = (
        ffmpeg
        .input(str(src))
        .filter('select', 'eq(n,{})'.format(frame_index))
        .output('pipe:', format='rawvideo', pix_fmt='rgb24', fps_mode='passthrough')
        .run(capture_stdout=True, quiet=True)
    )
    frame = np.frombuffer(out, np.uint8).reshape([h, w, 3])
    return frame

class VideoReader:
    """
    A (very) basic video iterator
    """
    def __init__(self, video_path, h=None, w=None, every_n_frames=1, start_frame=0, end_frame=None):
        """
        :param video_path: path to video
        :param target_resolution: (h, w) tuple of target resolution (must have common divisor with original resolution)
        """
        self.video_path = Path(video_path)
        self.h, self.w, self.fps, self.fc = get_video_info(video_path)
        self.th = None
        self.tw = None
        if h is not None and w is not None:
            self.th, self.tw = h, w
            if self.h % self.th != 0 or self.w % self.tw != 0:
                raise ValueError("target resolution must have common divisor with original resolution")
        self.start_frame = start_frame
        if end_frame is None:
            self.end_frame = self.fc + 1
        else:
            self.end_frame = end_frame
        self.every_n_frames = every_n_frames
        self.stream = (
            ffmpeg
            .input(str(self.video_path))
            .filter('select', 'between(n, {}, {})*not(mod(n,{}))'.format(self.start_frame, self.end_frame, self.every_n_frames))
            .output('pipe:', format='rawvideo', pix_fmt='rgb24', fps_mode='passthrough')
            .run_async(pipe_stdout=True, quiet=True)
        )

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self):
        if self.n < self.fc:
            in_bytes = self.stream.stdout.read(self.h * self.w * 3)
            if not in_bytes:  # should never happen
                self.stream.stdout.close()
                raise StopIteration
            frame = np.frombuffer(in_bytes, np.uint8).reshape([self.h, self.w, 3])
            if self.th is not None:
                frame = resize_images_naive(frame[None, ...], self.th, self.tw, mode="mean").squeeze()
            self.n += 1
            return frame
        else:
            in_bytes = self.stream.stdout.read(self.h * self.w * 3)
            assert not in_bytes  # sanity check, no bytes should be left
            self.stream.stdout.close()
            raise StopIteration



import numpy as np
import torch
import cv2
from .gsoup_io import save_image, save_images, load_images, load_image
from .image import interpolate_multi_channel, change_brightness
from pathlib import Path

def warp_image(p2c, cam_image, cam_h=None, cam_w=None, output_path=None):
    """
    given a 2D dense mapping between pixels from optical device 1 to optical device 2,
    warp an image from optical device 1 to optical device 2
    :param p2c: 2D dense mapping between pixels from optical device 1 to optical device 2
    :param cam_image: path to image from optical device 1
    :param cam_h: camera height, if not supplied assumes cam_image is in the correct dimensions in relation to p2c
    :param cam_w: camera width, if not supplied assumes cam_image is in the correct dimensions in relation to p2c
    :param output_path: path to save warped image to
    """
    if type(cam_image) == np.ndarray:
        unwarped = cam_image
    else:
        unwarped = load_image(cam_image, to_float=True, resize_wh=(cam_w, cam_h))
    #print(p2c.shape)
    #p2c = Image.open(Path(interpolated_p2c_path))
    if type(p2c) != np.ndarray:
        p2c = np.load(p2c)
    p2c = np.asarray(p2c)[:, :, :2]
    #p2c = p2c/255
    p2c[:, :, 0] = (p2c[:, :, 0] * 2) - 1
    p2c[:, :, 1] = (p2c[:, :, 1] * 2) - 1
    p2c[:, :, [1, 0]] = p2c[:, :, [0, 1]]
    # p2c = np.round(p2c).astype(np.int32)
    grid = torch.tensor(p2c.astype(np.float32)).unsqueeze(0)
    input = torch.tensor(unwarped).permute(2, 0, 1).unsqueeze(0)
    warped = torch.nn.functional.grid_sample(input, grid).squeeze().permute(1, 2, 0).numpy()
    warped_int = (warped * 255).astype(np.uint8)
    if output_path is not None:
        output_path.parent.mkdir(exist_ok=True, parents=True)
        save_image(warped_int, output_path)
    return warped_int

def generate_gray_code(height, width, step, output_dir=None):
    """
    generate gray code patterns for structured light scanning
    :param height: height of the pattern
    :param width: width of the pattern
    :param step: step size of the pattern (e.g. 2 means the patterns are half the resolution)
    :param output_dir: directory to save the patterns to
    :return: list of patterns
    """
    gc_height = int((height-1)/step)+1
    gc_width = int((width-1)/step)+1
    graycode = cv2.structured_light_GrayCodePattern.create(gc_width, gc_height)
    patterns = graycode.generate()[1]
    # decrease pattern resolution
    exp_patterns = []
    for pat in patterns:
        img = np.zeros((height, width), np.uint8)
        for y in range(height):
            for x in range(width):
                img[y, x] = pat[int(y/step), int(x/step)]
        exp_patterns.append(img)
    exp_patterns.append(255*np.ones((height, width), np.uint8))  # white
    exp_patterns.append(np.zeros((height, width), np.uint8))    # black
    exp_patterns = np.stack(exp_patterns)
    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(exist_ok=True, parents=True)
        file_names = ["pattern_{:02d}.png".format(i) for i in range(len(exp_patterns))]
        save_images(exp_patterns[..., None], output_dir, file_names=file_names)
    return exp_patterns

def pix2pix_correspondence(proj_width, proj_height, step, captures,
                           BLACKTHR = 2, WHITETHR = 30, output_dir=None, debug=False):
    """
    finds dense pixel to pixel pix2pix_correspondence between a projector and a camera
    note: assumes gray code patterns used for projections were generated using generate_gray_code
    :param proj_width: width of projector
    :param proj_height: height of projector
    :param step: step factor used for gray code patterns (e.g. 2 means half resolution)
    :param captures: the actual n x cam_height x cam_width x 3 captured images, or a directory containing the images
    :param BLACKTHR: threshold for black pixels
    :param WHITETHR: threshold for white pixels
    :param output_dir: directory to save results to
    :param debug: if True, saves debug images
    """
    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)
    # prep decoder
    gc_width = int((proj_width-1)/step)+1
    gc_height = int((proj_height-1)/step)+1
    graycode = cv2.structured_light_GrayCodePattern.create(gc_width, gc_height)
    graycode.setBlackThreshold(BLACKTHR)
    graycode.setWhiteThreshold(WHITETHR)
    correct_pattern_amount = graycode.getNumberOfPatternImages() + 2
    if type(captures) != np.ndarray:
        captures = load_images(captures)
    captures = captures.mean(axis=-1).astype(np.uint8)  # convert to grayscale
    if len(captures) != correct_pattern_amount:
        raise ValueError('Number of images is not right (right number is {})'.format(correct_pattern_amount))
    imgs = list(captures)
    black = imgs.pop()
    white = imgs.pop()
    cam_height = white.shape[0]
    cam_width = white.shape[1]
    if debug:
        diff_pic = white.astype(np.uint64) - black.astype(np.uint64)
        diff_pic[diff_pic < 0] = 0
        save_image(diff_pic[..., None].astype(np.uint8), Path(output_dir, "white_black_diff.png"))
        print('camera image size :', white.shape)
    # initialize
    viz_c2p = np.zeros((cam_height, cam_width, 3), np.float32)
    ragged_p2c = np.empty((proj_height, proj_width), dtype=object)
    for i in np.ndindex(ragged_p2c.shape): ragged_p2c[i] = []
    missing_values_c2p = np.ones((cam_height, cam_width), np.bool8)
    # c2p
    c2p_list = [] # [((cam x, y), (proj x, y))]
    for y in range(cam_height):
        for x in range(cam_width):
            if int(white[y, x]) - int(black[y, x]) <= BLACKTHR:  # background
                continue
            err, proj_pix = graycode.getProjPixel(imgs, x, y)
            if not err:
                fixed_pix = step*(proj_pix[0]+0.5), step*(proj_pix[1]+0.5)  # x, y
                ragged_p2c[int(fixed_pix[1]), int(fixed_pix[0])].append([y, x])
                viz_c2p[y, x, :] = [fixed_pix[1] / proj_height, fixed_pix[0] / proj_width, 0.0]
                missing_values_c2p[y, x] = False
                c2p_list.append(((x, y), fixed_pix))
    # p2c
    total_size = ragged_p2c.size
    counter = 0
    for i in range(proj_height):
        for j in range(proj_width):
            print("{} / {}".format(counter, total_size), end="\r", flush=True)
            val = np.mean(ragged_p2c[i, j], dtype=np.float32, axis=0)
            if np.isnan(val).any():
                val = np.zeros(2)
            else:
                val = np.round(val).astype(np.int32)
            ragged_p2c[i, j] = val
            counter += 1
    viz_p2c = np.vstack(ragged_p2c.reshape(-1)).reshape(proj_height, proj_width, 2).astype(np.uint32)
    viz_p2c = viz_p2c / np.array([cam_height, cam_width], dtype=np.float32)[None, None, :]
    viz_p2c = np.concatenate((viz_p2c, np.zeros_like(viz_p2c)[:, :, 0:1]), axis=-1)
    missing_values_p2c = (viz_p2c == 0).all(axis=-1)
    # interpolate missing values
    interpolated_c2p = interpolate_multi_channel(viz_c2p, missing_values_c2p)
    interpolated_p2c = interpolate_multi_channel(viz_p2c, missing_values_p2c)
    # save results
    if output_dir is not None:
        np.save(Path(output_dir, "c2p.npy"), interpolated_c2p) 
        np.save(Path(output_dir, "p2c.npy"), interpolated_p2c) 
        if debug:
            save_image(interpolated_c2p, Path(output_dir, "interpolated_c2p.png"))
            save_image(interpolated_p2c, Path(output_dir, "interpolated_p2c.png"))
            save_image(viz_c2p, Path(output_dir, "c2p.png"))
            save_image(viz_p2c, Path(output_dir, "p2c.png"))
            print('Amount of c2p correspondences :', len(c2p_list))
    return interpolated_c2p, interpolated_p2c

def naive_color_compensate(target_image, all_white_image, all_black_image, cam_width, cam_height, brightness_decrease=-127, output_dir=None, debug=False):
    """
    color compensate a projected image such that it appears closer to a target image from the perspective of a camera
    loosly based on ***insert citation***
    :param target_image the desired image from the perspective of the camera
    :param all_white_image a picture taken by camera when projector had all pixels fully on
    :param all_black_image a picture taken by camera when projector had all pixels fully off
    :param cam_width camera image width
    :param cam_height camera image height
    :param brightness_decrease a hyper parameter controlling how much the total brightness is decreased. without this, the result is saturated because of dividing by small numbers
    :output_dir if passed, result will be saved here
    :debug if true, will save debug info into outputdir 
    """
    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)
    target_image = load_image(target_image, to_float=True, resize_wh=(cam_width, cam_height))
    target_image = change_brightness(target_image, brightness_decrease)
    if debug:
        save_image(target_image, "target_decrease_brightness.png") 
    #unwarped_image = np.power(unwarped_image, -2.2)
    compensated = (target_image - all_black_image) / all_white_image
    compensated = np.power(compensated, (1/2.2))
    compensated = np.nan_to_num(compensated, nan=0.0, posinf=0.0, neginf=0.0)
    compensated = np.clip(compensated, 0, 1)
    if output_dir:
        save_image(compensated, "compensated.png")
    return compensated

def calibrate_procam(proj_height, proj_width, graycode_step, capture_dir, 
                     chess_vert=10, chess_hori=7,
                     black_thr=40, white_thr=5, chess_block_size=10.0, verbose=True):
    """
    calibrates a projection-camera pair
    :param proj_height projector pixel height
    :param proj_width projector pixel width
    :param chess_vert number of cross points of chessboard in vertical direction (not including the border, i.e. internal corners)
    :param chess_hori number of cross points of chessboard in horizontal direction (not including the border, i.e. internal corners)
    :param graycode_step factor used to downsample the graycode images (see generate_gray_code)
    :param capture_dir directory containing the captured images when gray code patterns were projected, assumes structure as follows:
        capture_dir
            - folder1
                - 0000.png
                - 0001.png
                - ...
            - folder2
                - 0000.png
                - ...
    :param black_thr threshold for detecting black pixels in the chessboard
    :param white_thr threshold for detecting white pixels in the chessboard
    :param chess_block_size size of blocks of chessboard in real world in any unit of measurement (will only effect the translation componenet between camera and projector)
    :param verbose if true, will print out the calibration results
    :return camera intrinsics, camera extrinsics, projector intrinsics, projector extrinsics, cam_proj_rmat, cam_proj_tvec
    """
    proj_shape = (proj_height, proj_width)
    chess_shape = (chess_vert, chess_hori)
    gc_step = graycode_step
    capture_dir = Path(capture_dir)
    if not capture_dir.exists():
        raise FileNotFoundError("capture_dir was not found")
    dirnames = sorted(capture_dir.glob('*'))
    if len(dirnames) == 0:
        raise FileNotFoundError("capture_dir contains no subfolders")
    used_dirnames = []
    gc_fname_lists = []
    for dname in dirnames:
        gc_fnames = sorted(dname.glob('*'))
        if len(gc_fnames) == 0:
            continue
        used_dirnames.append(str(dname))
        gc_fname_lists.append([str(x) for x in gc_fnames])
    dirnames = used_dirnames
    objps = np.zeros((chess_shape[0]*chess_shape[1], 3), np.float32)
    objps[:, :2] = chess_block_size * np.mgrid[0:chess_shape[0], 0:chess_shape[1]].T.reshape(-1, 2)
    gc_height = int((proj_shape[0]-1)/gc_step)+1
    gc_width = int((proj_shape[1]-1)/gc_step)+1
    graycode = cv2.structured_light_GrayCodePattern.create(gc_width, gc_height)
    graycode.setBlackThreshold(black_thr)
    graycode.setWhiteThreshold(white_thr)
    cam_shape = load_image(gc_fname_lists[0][0], as_grayscale=True).shape
    patch_size_half = int(np.ceil(cam_shape[1] / 180))
    # print('  patch size :', patch_size_half * 2 + 1)

    cam_corners_list = []
    cam_objps_list = []
    cam_corners_list2 = []
    proj_objps_list = []
    proj_corners_list = []
    for dname, gc_filenames in zip(dirnames, gc_fname_lists):
        if len(gc_filenames) != graycode.getNumberOfPatternImages() + 2:
            raise ValueError("invalid number of images in " + dname)

        imgs = []
        for fname in gc_filenames:
            img = load_image(fname, as_grayscale=True)
            if cam_shape != img.shape:
                raise ValueError("image size of {} does not match other images".format(fname))
            imgs.append(img)
        black_img = imgs.pop()
        white_img = imgs.pop()

        res, cam_corners = cv2.findChessboardCorners(white_img, chess_shape)
        if not res:
            raise RuntimeError("chessboard was not found in {}".format(gc_filenames[-2]))
        cam_objps_list.append(objps)
        cam_corners_list.append(cam_corners)

        proj_objps = []
        proj_corners = []
        cam_corners2 = []
        # viz_proj_points = np.zeros(proj_shape, np.uint8)
        for corner, objp in zip(cam_corners, objps):
            c_x = int(round(corner[0][0]))
            c_y = int(round(corner[0][1]))
            src_points = []
            dst_points = []
            for dx in range(-patch_size_half, patch_size_half + 1):
                for dy in range(-patch_size_half, patch_size_half + 1):
                    x = c_x + dx
                    y = c_y + dy
                    if int(white_img[y, x]) - int(black_img[y, x]) <= black_thr:
                        continue
                    err, proj_pix = graycode.getProjPixel(imgs, x, y)
                    if not err:
                        src_points.append((x, y))
                        dst_points.append(gc_step*np.array(proj_pix))
            if len(src_points) < patch_size_half**2:
                if verbose:
                    print('corner {}, {} was skiped because too few decoded pixels found (check your images and threasholds)'.format(c_x, c_y))
                continue
            h_mat, inliers = cv2.findHomography(
                np.array(src_points), np.array(dst_points))
            point = h_mat@np.array([corner[0][0], corner[0][1], 1]).transpose()
            point_pix = point[0:2]/point[2]
            proj_objps.append(objp)
            proj_corners.append([point_pix])
            cam_corners2.append(corner)
            # viz_proj_points[int(round(point_pix[1])),
            #                 int(round(point_pix[0]))] = 255
        if len(proj_corners) < 3:
            raise RuntimeError("too few corners were found in {} (less than 3)".format(dname))
        proj_objps_list.append(np.float32(proj_objps))
        proj_corners_list.append(np.float32(proj_corners))
        cam_corners_list2.append(np.float32(cam_corners2))
        # cv2.imwrite('visualize_corners_projector_' +
        #             str(cnt) + '.png', viz_proj_points)
        # cnt += 1

    # Initial solution of camera's intrinsic parameters
    ret, cam_int, cam_dist, cam_rvecs, cam_tvecs = cv2.calibrateCamera(
        cam_objps_list, cam_corners_list, cam_shape, None, None, None, None)
    if verbose:
        print('Initial camera intrinsic parameters: {}'.format(cam_int))
        print('Initial camera distortion parameters: {}'.format(cam_dist))
        print('Initial camera RMS: {}'.format(ret))

    # Initial solution of projector's parameters
    ret, proj_int, proj_dist, proj_rvecs, proj_tvecs = cv2.calibrateCamera(
        proj_objps_list, proj_corners_list, proj_shape, None, None, None, None)
    if verbose:
        print('Initial projector intrinsic parameters: {}'.format(proj_int))
        print('Initial projector distortion parameters: {}'.format(proj_dist))
        print('Initial projector RMS: {}'.format(ret))

    # Stereo calibration for final solution
    ret, cam_int, cam_dist, proj_int, proj_dist, cam_proj_rmat, cam_proj_tvec, E, F = cv2.stereoCalibrate(
        proj_objps_list, cam_corners_list2, proj_corners_list, cam_int, cam_dist, proj_int, proj_dist, None)
    
    if verbose:
        print('RMS: {}'.format(ret))
        print('Camera intrinsic parameters: {}'.format(cam_int))
        print('Camera distortion parameters: {}'.format(cam_dist))
        print('Projector intrinsic parameters: {}'.format(proj_int))
        print('Projector distortion parameters: {}'.format(proj_dist))
        print('Rotation matrix / translation vector from camera to projector (cam2proj transform): {}, {}'.format(cam_proj_rmat, cam_proj_tvec))
    return cam_int, cam_dist, proj_int, proj_dist, cam_proj_rmat, cam_proj_tvec


import numpy as np
from gsoup.viewer import gviewer
from gsoup import structures

"""
a few example functions for using the viewer
"""

#### globals 
ui_float = 0.0
poses = None
meshes_v = None
meshes_f = None
meshes_attribute = None
#### globals

def meshes_slider_callback():
    global ui_float, meshes_v, meshes_f
    changed, ui_float = gviewer.psim.SliderFloat("step", ui_float, v_min=0, v_max=len(meshes_v))
    if changed:
        if int(ui_float) >= len(meshes_v):
            ui_float = len(meshes_v)-1
        gviewer.register_mesh("mesh", meshes_v[int(ui_float)],
                    meshes_f[int(ui_float)],
                    v_vertices=meshes_attribute[int(ui_float)], edge_width=1.0)

def poses_slider_callback():
    global ui_float, poses
    edge_rad = 0.0005
    changed, ui_float = gviewer.psim.SliderFloat("step", ui_float, v_min=0, v_max=len(poses))
    if changed and poses is not None:
        if int(ui_float) >= len(poses):
            ui_float = len(poses)-1
        v_tot, e_tot, c_tot = gviewer.register_camera("poses", poses[int(ui_float)], edge_rad, group=True)

def meshes_slide_view(v, f, v_attribute):
    """
    given some vertices tXVx3 and faces tXFx3
    show the mesh as it changes through time t and allow scrolling through using a slider.
    """
    global meshes_v, meshes_f, meshes_attribute
    gviewer.init()
    meshes_v = v
    meshes_f = f
    meshes_attribute = v_attribute
    gviewer.ps.set_user_callback(meshes_slider_callback)
    gviewer.ps.set_up_dir("z_up")
    gviewer.register_mesh("mesh", meshes_v[0], meshes_f[0],
                v_vertices=meshes_attribute[0], edge_width=1.0)
    gviewer.show()
        
def poses_slide_view(camera_poses):
    """
    given a tensor of t x b x 4 x 4 camera poses, where t is time axis (or step number), b is batch axis, and 4x4 is the camera pose matrix,
    show the batch of poses and allow scrolling through the time axis using a slider.
    """
    if camera_poses.ndim != 4:
        raise ValueError("camera_poses must be t x b x 4 x 4")
    if camera_poses.shape[2] != 4 or camera_poses.shape[3] != 4:
        raise ValueError("camera_poses must be t x b x 4 x 4")
    global poses
    poses = camera_poses
    gviewer.init()
    gviewer.ps.set_user_callback(poses_slider_callback)
    gviewer.ps.set_up_dir("z_up")
    gviewer.ps.set_ground_plane_mode("shadow_only")
    edge_rad = 0.0005
    point_rad = 0.002
    v_aabb, e_aabb, c_aabb = structures.get_aabb_coords()
    ps_net = gviewer.ps.register_curve_network("aabb", v_aabb, e_aabb, radius=edge_rad)
    ps_net.add_color_quantity("color", c_aabb, defined_on='edges', enabled=True)
    gviewer.register_pointcloud("center_of_world", np.zeros((1, 3)), c=np.array([1., 1., 1.])[None, :], radius=0.005, mode="sphere")
    v_tot, e_tot, c_tot = gviewer.register_camera("poses_orig", poses[0], edge_rad, group=True, alpha=0.3)
    gviewer.ps.show()

def poses_static_view(camera_poses=None, meshes=None, pointclouds=None, group_cameras=True):
    """
    visualizes a camera setup
    :param camera_pose: (n, 4, 4) np array of camera to world transforms
    :param meshes: list of (v, f) tuples
    :param pointclouds: list of v
    :param group_cameras: if true, accelerates view but groups camera as a single object
    :return:
    """
    gviewer.init(height=512, width=512)
    gviewer.ps.set_up_dir("z_up")
    edge_rad = 0.0005
    v_aabb, e_aabb, c_aabb = structures.get_aabb_coords()
    ps_net = gviewer.ps.register_curve_network("aabb", v_aabb, e_aabb, radius=edge_rad)
    ps_net.add_color_quantity("color", c_aabb, defined_on='edges', enabled=True)
    gviewer.register_pointcloud("center_of_world", np.zeros((1, 3)), c=np.array([1., 1., 1.])[None, :], radius=0.005, mode="sphere")
    if camera_poses is not None:
        v_tot, e_tot, c_tot = gviewer.register_camera("cameras", camera_poses, edge_rad, group_cameras)
    if meshes is not None:
        for i, mesh in enumerate(meshes):
            gviewer.register_mesh("mesh_{}".format(i), mesh[0], mesh[1], transparency=0.5)
    if pointclouds is not None:
        for i, pointcloud in enumerate(pointclouds):
            gviewer.register_pointcloud("pc_{}".format(i), pointcloud, radius=1e-4)
    gviewer.show()


import torch
import numpy as np
from PIL import Image

def to_hom(x):
    """
    converts a vector to homogeneous coordinates
    :param x: nxc numpy array
    :return: nxc+1 numpy array
    """
    if type(x) == torch.Tensor:
        if x.ndim == 1:
            return torch.cat((x, torch.ones(1, device=x.device)))
        else:
            return torch.cat((x, torch.ones(x.shape[0], 1, device=x.device)), dim=-1)
    elif type(x) == np.ndarray:
        if x.ndim == 1:
            return np.concatenate((x, np.array([1], dtype=x.dtype)))
        else:
            return np.concatenate((x, np.ones((x.shape[0], 1), dtype=x.dtype)), axis=-1)
    else:
        raise ValueError("x must be torch.Tensor or np.ndarray")

def homogenize(x, keepdim=False):
    """
    normalizes a homogeneous vector by dividing by the last coordinate
    :param x: nx3 numpy array
    :return: nx4 numpy array
    """
    x = (x / x[..., -1:])
    if not keepdim:
        x = x[..., :-1]
    return x

def normalize(x, eps=1e-7):
    if type(x) == torch.Tensor:
        return x / (torch.norm(x, dim=-1, keepdim=True) + eps)
    elif type(x) == np.ndarray:
        return x / (np.linalg.norm(x, axis=-1, keepdims=True) + eps)


def broadcast_batch(*args):
    """
    broadcast a list of arrays to the same shape on the batch dimension
    assumes first dimension is batch unless ndim = 1 for all inputs (but then does not broadcast)
    :param args: list of arrays
    :return: list of arrays with the same shape
    """
    shapes = [a.shape for a in args]
    ndims = np.array([a.ndim for a in args])
    new_args = []
    if (ndims <= 1).all():
        for a in args:
            new_args.append(a[None, :])
        return new_args
    elif ndims.min() == 1:
        raise ValueError("cannot broadcast 1d and nd arrays")
    else:
        batch_dim = np.array([s[0] for s in shapes]).max()
        for i, a in enumerate(args):
            if a.shape[0] == batch_dim:
                new_args.append(a)
            else:
                new_args.append(np.broadcast_to(a, (batch_dim, *a.shape[1:])))
    return new_args

def compose_rt(R: np.array, t: np.array):
    """
    composes a n x 3 x 4 numpy array from rotation and translation.
    will broadcast upon batch dimension if necessary.
    :param R: nx3x3 numpy array
    :param t: nx3 numpy array
    :return: n x 3 x 4 composition of the rotation and translation
    """
    RR, tt = broadcast_batch(R, t)
    return np.concatenate((RR, tt[:, :, None]), axis=-1)

def to_44(mat: np.array):
    """
    converts a 3x4 to a 4x4 matrix by concatenating 0 0 0 1
    :param mat: dimsx3x4 numpy array (dims can be any number of dims including 0)
    :return: dimsx4x4 numpy array
    """
    if mat.shape[-2:] != (3, 4):
        raise ValueError("mat must be 3x4")
    to_cat = np.broadcast_to(np.array([0, 0, 0, 1]), (*mat.shape[:-2], 1, 4))
    new_mat = np.concatenate((mat, to_cat), axis=-2)
    return new_mat

def to_34(mat: np.array):
    """
    converts a 4x4 to a 3x4 matrix by removeing the last row
    :param mat: 4x4 numpy array
    :return: 3x4 numpy array
    """
    if mat.ndim == 3:
        if mat.shape[1:] != (4, 4):
            raise ValueError("mat must be 4x4")
        return mat[:, :-1, :]
    else:
        if mat.shape != (4, 4):
            raise ValueError("mat must be 4x4")
        return mat[:-1, :]

def look_at_np(from_, to_, up_, openGL=False):
    """
    returns a batch of look_at transforms 4x4 (camera->world, the inverse of a ModelView matrix)
    will broadcast upon batch dimension if necessary.
    :param from_: n x 3 from vectors in world space
    :param to_: n x 3 at vector in world space
    :param up_: n x 3 up vector in world space
    :param is_openGL: if True, output will be in opengl coordinates (z backward, y up) otherwise (z forward, y down)
    :return: n x 4 x 4 transformation matrices (camera->world, the inverse of a ModelView matrix)
    """
    from_, to_, up_ = broadcast_batch(from_, to_, up_)
    forward = to_ - from_
    forward = forward / np.linalg.norm(forward, axis=-1, keepdims=True)
    right = np.cross(forward, up_)
    right = right / np.linalg.norm(right, axis=-1, keepdims=True)
    up = np.cross(forward, right)
    up = up / np.linalg.norm(up, axis=-1, keepdims=True)
    rot = np.concatenate((right[..., None], up[..., None], forward[..., None]), axis=-1)
    c2w = np.concatenate((rot, from_[..., None]), axis=-1)
    c2w = to_44(c2w)
    if openGL:
        c2w[:, :, 1] *= -1
        c2w[:, :, 2] *= -1
    return c2w

def look_at_torch(
        eye:torch.Tensor, #3
        at:torch.Tensor, #3
        up:torch.Tensor, #3
        device:torch.device,
        openGL:bool=False
    ) -> torch.Tensor: #4,4
    """
    creates a lookat transform matrix (OpenCV convention)
    :param eye: where the camera is
    :param at: where the camera is looking
    :param up: the up vector of world space
    :param device: the device to put the matrix on
    :return: 4x4 lookat transform matrix
    """
    if openGL:
        z = (eye - at).type(torch.float32).to(device)
    else:
        z = (at - eye).type(torch.float32).to(device)
    z /= torch.norm(z)
    x = torch.cross(up, z).type(torch.float32).to(device)
    x /= torch.norm(x)
    y = torch.cross(z, x).type(torch.float32).to(device)
    y /= torch.norm(y)
    T = torch.eye(4, device=device)
    T[:3,:3] = torch.stack([x,y,z],dim=1)
    T[:3,3] = eye
    return T

def orthographic_projection(l, r, b, t, n, f):
    """
    creates an orthographic projection matrix (OpenGL convention)
    :param l: left
    :param r: right
    :param b: bottom
    :param t: top
    :param n: near
    :param f: far
    :return: 4x4 orthographic projection matrix
    """
    dx = r - l
    dy = t - b
    dz = f - n
    rx = -(r + l) / (r - l)
    ry = -(t + b) / (t - b)
    rz = -(f + n) / (f - n)
    return np.array([[2.0/dx,0,0,rx],
                      [0,2.0/dy,0,ry],
                      [0,0,-2.0/dz,rz],
                      [0,0,0,1]])

def perspective_projection(fovy=45, aspect=1.0, n=0.1, f=100.0):
    """
    creates a perspective projection matrix (OpenGL convention)
    :param fovy: field of view in y direction
    :param aspect: aspect ratio
    :param n: near plane
    :param f: far plane
    :return: 4x4 projection matrix
    """
    s = 1.0/np.tan(np.deg2rad(fovy)/2.0)
    sx, sy = s / aspect, s
    zz = (f+n)/(n-f)
    zw = 2*f*n/(n-f)
    return np.array([[sx,0,0,0],
                      [0,sy,0,0],
                      [0,0,zz,zw],
                      [0,0,-1,0]])

def frustum_projection(x0, x1, y0, y1, z0, z1):
    """
    creates a projection matrix from a frustum (OpenGL convention)
    :param x0: left
    :param x1: right
    :param y0: bottom
    :param y1: top
    :param z0: near
    :param z1: far
    :return: 4x4 projection matrix
    """
    a = (x1+x0)/(x1-x0)
    b = (y1+y0)/(y1-y0)
    c = -(z1+z0)/(z1-z0)
    d = -2*z1*z0/(z1-z0)
    sx = 2*z0/(x1-x0)
    sy = 2*z0/(y1-y0)
    return np.array([[sx, 0, a, 0],
                      [ 0,sy, b, 0],
                      [ 0, 0, c, d],
                      [ 0, 0,-1, 0]])

def opengl_project_from_opencv_intrinsics(opencv_intrinsics, width, height, near=0.1, far=100.0):
    """
    given a matrix K from opencv, returns the corresponding projection matrix for opengl ("Eye/Camera/View space -> Clip Space")
    :param opencv_project: 3x3 projection matrix from opencv
    :param width: width of the image
    :param height: height of the image
    :param near: near plane
    :param far: far plane
    :return: 4x4 projection matrix for opengl (note: column major)
    """
    fx = opencv_intrinsics[0, 0]
    fy = opencv_intrinsics[1, 1]
    cx = opencv_intrinsics[0, 2]
    cy = opencv_intrinsics[1, 2]
    opengl_mtx = np.array([[2*fx/width, 0.0, (width - 2*cx)/width, 0.0],
                           [0.0, -2*fy/height, (height - 2*cy)/height, 0.0],
                           [0.0, 0.0, (-far - near) / (far - near), -2.0*far*near/(far-near)],
                           [0.0, 0.0, -1.0, 0.0]])
    return opengl_mtx

def opengl_c2w_to_opencv_c2w(opengl_transforms):
    """
    given a modelview matrix (World space->Eye/Camera/View space) where z is backward and y is up,
    converts its coordinate system to opencv convention (z forward, y down)
    :param opengl_transforms: 4x4 modelview matrix or batch of 4x4 modelview matrices
    :return: 4x4 modelview matrix in opencv convention
    """
    if opengl_transforms.ndim == 2:
        my_transforms = opengl_transforms[None, ...]
    else:
        my_transforms = opengl_transforms
    if my_transforms.shape[1:] != (4, 4):
        raise ValueError("transform must be 4x4 or batch of 4x4")
    my_transforms[:, :, 1] *= -1
    my_transforms[:, :, 2] *= -1
    return my_transforms.reshape(opengl_transforms.shape)

def opencv_c2w_to_opengl_c2w(opencv_transform):
    """
    converts coordinates of "vision" (opencv) to opengl coordinates by flipping y and z axes
    """
    return opengl_c2w_to_opencv_c2w(opencv_transform)

def create_random_cameras_on_unit_sphere(n, r, device="cuda"):
    """
    creates a batch of world2view ("ModelView" matrix) and view2clip ("Projection" matrix) transforms on a unit sphere looking at the center
    :param n: number of cameras
    :param r: radius of the sphere
    :param device: device to put the tensors on
    :return: world2view, view2clip
    """
    locs = torch.randn((n, 3), device=device)
    locs = torch.nn.functional.normalize(locs, dim=1, eps=1e-6)
    locs = locs * r
    matrices = torch.empty((n, 4, 4), dtype=torch.float32, device=device)
    for i in range(len(locs)):
        matrices[i] = look_at_torch(locs[i],
                                       torch.zeros(3, dtype=torch.float32, device=device),
                                       torch.tensor([0.,1.,0.], device=device),
                                       device=device)
    v2w = matrices  # c2w
    w2v = torch.inverse(v2w)
    v2c = torch.tensor(perspective_projection(), dtype=torch.float32, device=device)
    return w2v, v2c

def to_np(arr: torch.Tensor):
    """
    converts a tensor to numpy array
    :param arr: tensor
    :return: numpy array
    """
    if type(arr) == torch.Tensor:
        return arr.detach().cpu().numpy()
    elif type(arr) == np.ndarray:
        return arr
    elif type(arr) == Image.Image:
        return np.array(arr)
    return arr.detach().cpu().numpy()

def to_numpy(arr: torch.Tensor):
    """
    converts a tensor to numpy array
    :param arr: tensor
    :return: numpy array
    """
    return to_np(arr)

def to_torch(arr: np.array, device="cpu", dtype=None):
    """
    converts a numpy array to a torch tensor
    :param arr: numpy array
    :param dtype: dtype of the tensor
    :param device: device to put the tensor on
    :return: torch tensor
    """
    if dtype is None:
        return torch.tensor(arr, device=device)
    else:
        return torch.tensor(arr, dtype=dtype, device=device)

def to_8b(x: np.array, clip=True):
    """
    convert a numpy (float, double) array to 8 bit
    """
    if x.dtype == np.float32 or x.dtype == np.float64:
        if clip:
            x = np.clip(x, 0, 1)
        return (255 * x).astype(np.uint8)
    elif x.dtype == np.uint8:
        return x

def to_float(x: np.array, clip=True):
    """
    convert a numpy (8bit) array to float
    """
    if x.dtype == np.uint8:
        return x.astype(np.float32) / 255
    elif x.dtype == np.float32:
        if clip:
            x = np.clip(x, 0, 1)
        return x
    else:
        raise ValueError("unsupported dtype")

def to_PIL(x: np.array):
    """
    convert a numpy float array to a PIL image
    """
    if x.ndim == 3:
        return Image.fromarray(to_8b(x))
    elif x.ndim == 2:
        return Image.fromarray(to_8b(x[:, :, None]), mode="L")
    else:
        raise ValueError("unsupported array dimensions")

def translate(t):
    """
    creates a translation matrix from a translation vector
    :param t: translation vector or batch of translation vectors
    :return: 4x4 translation matrix
    """
    if t.shape[-1] != 3:
        raise ValueError("translation vector must be 3d")
    if t.ndim == 1:
        t = t[None, :]
    mat = np.concatenate((np.eye(3)[None, :, :], t[:, :, None]), axis=-1)
    return to_44(mat)

def scale(s):
    """
    creates a scaling matrix from a scaling vector
    :param s: scaling vector or batch of scaling vectors
    :return: nx4x4 scaling matrix
    """
    if s.shape[-1] != 3:
        raise ValueError("translation vector must be 3d")
    if s.ndim == 1:
        s = s[None, :]
    mat = np.diag(s)
    return to_44(mat)

def sincos(a):
    a = np.deg2rad(a)
    return np.sin(a), np.cos(a)

def rotate(a, r):
    """
    creates a rotation matrix from an angle and an axis
    """
    s, c = sincos(a)
    r = normalize(r)
    nc = 1 - c
    x, y, z = r
    return np.array([[x*x*nc +   c, x*y*nc - z*s, x*z*nc + y*s, 0],
                      [y*x*nc + z*s, y*y*nc +   c, y*z*nc - x*s, 0],
                      [x*z*nc - y*s, y*z*nc + x*s, z*z*nc +   c, 0],
                      [           0,            0,            0, 1]])

def rotx(a):
    """
    creates a rotation matrix around the x axis
    """
    s, c = sincos(a)
    return np.array([[1,0,0,0],
                      [0,c,-s,0],
                      [0,s,c,0],
                      [0,0,0,1]])

def roty(a):
    """
    creates a rotation matrix around the y axis
    """
    s, c = sincos(a)
    return np.array([[c,0,s,0],
                      [0,1,0,0],
                      [-s,0,c,0],
                      [0,0,0,1]])

def rotz(a):
    """
    creates a rotation matrix around the z axis
    """
    s, c = sincos(a)
    return np.array([[c,-s,0,0],
                      [s,c,0,0],
                      [0,0,1,0],
                      [0,0,0,1]])

def map_range(x, in_min, in_max, out_min, out_max):
    """
    given an input and a range, maps it to a new range
    """
    if not in_min <= x <= in_max:
        raise ValueError("input must be inside range ({} - {})".format(in_min, in_max))
    return int((x-in_min) * (out_max-out_min) / (in_max-in_min) + out_min)


import numpy as np
from . import structures
from .core import to_hom, homogenize
from .gsoup_io import write_to_json

class PolyscopeStub:
    def __getattr__(self, name):
        def wrapper(*args, **kwargs):
            print("{} was called".format(name))
        return wrapper

class gviewer():
    """
    wrapper class for polyscope, invoked as a singelton ("from gsoup.viewer import gviewer")
    offers a stub for when polyscope is not installed, window resizing and a few other things
    """
    def __init__(self):
        self.is_init = False

    def init(self, up_dir="z_up", look_at=None, ground_mode="shadow_only", projection_mode="perspective", width=None, height=None):
        """
        given a ps instance, intializes it
        :param up_dir: up direction
        :param look_at: where to look at (3 tuples eye, at ,up)
        :param ground_mode: ground render mode
        :param projection_mode: projection_mode
        """
        if width is not None:
            if height is None:
                height = width
            poly_dict = {"windowHeight": width, "windowPosX": 50, "windowPosY": 50, "windowWidth": height}
            write_to_json(poly_dict, "./.polyscope.ini")
        try:
            import polyscope as ps
            import polyscope.imgui as psim
            ps.init()
        except RuntimeError:
            ps = PolyscopeStub()
            psim = PolyscopeStub()
        self.ps = ps
        self.psim = psim
        self.ps.set_up_dir(up_dir)
        if look_at is None:
            eye, at, up = (-1., 0.75, -1.), (0., 0., 0.), (0., 1., 0.)
        else:
            eye, at, up = look_at
        self.ps.look_at_dir(eye, at, up)
        self.ps.set_ground_plane_mode(ground_mode)
        self.ps.set_view_projection_mode(projection_mode)  # ps.set_view_projection_mode("orthographic")
        self.is_init = True

    def register_pointcloud(self, name, points, c=None, s=None, v=None, radius=1e-2, mode="quad"):
        """
        register a point cloud to polyscope
        """
        ps_pointcloud = self.ps.register_point_cloud(name, points,
                                                radius=radius,
                                                point_render_mode=mode)  # color=np.array([0.5, 0.1, 0.3])  # sphere / quad
        if c is not None:
            ps_pointcloud.add_color_quantity("colors", c, enabled=True)
        if s is not None:
            s = (s - np.min(s)) / (np.max(s) - np.min(s))
            ps_pointcloud.add_scalar_quantity("scalar_value", s,
                                            enabled=True,
                                            vminmax=(0., 1.),
                                            cmap="reds")
        if v is not None:
            # v /= (np.linalg.norm(v, axis=-1) + 1e-6)
            ps_pointcloud.add_vector_quantity("vecs", v,
                                            enabled=True,
                                            radius=0.002,
                                            length=0.1,
                                            color=(0.2, 0.5, 0.5))
        return ps_pointcloud

    def register_camera(self, name, poses, edge_rad, group=True, alpha=1.0):
        """
        register a camera structure to polyscope
        """
        v_cam, e_cam, c_cam = structures.get_camera_coords()
        v_tot = []
        e_tot = []
        c_tot = []
        for i, pose in enumerate(poses):
            v = homogenize((pose @ to_hom(v_cam).T).T)
            if group:
                v_tot.append(v)
                c_tot.append(c_cam)
                e_tot.append(e_cam + (i * len(v_cam)))
            else:
                ps_net = self.ps.register_curve_network("{}_{}".format(name, i), v, e_cam, radius=edge_rad)
                ps_net.add_color_quantity("color", c_cam, defined_on='edges', enabled=True)
                ps_net.set_transparency(alpha)
        if group:
            v_tot = np.array(v_tot).reshape(-1, 3)
            e_tot = np.array(e_tot).reshape(-1, 2)
            c_tot = np.array(c_tot).reshape(-1, 3)
            ps_net = self.ps.register_curve_network(name, v_tot, e_tot, radius=edge_rad)
            ps_net.add_color_quantity("color", c_tot, defined_on='edges', enabled=True)
        ps_net.set_transparency(alpha)
        return v_tot, e_tot, c_tot

    def register_mesh(self, name, v, f,
                    transparency=1.0, edge_width=0., color=[0.5, 0.5, 0.5], smooth_shade=True,
                    c_vertices=None, c_faces=None, s_faces=None, v_vertices=None, cmap=None, enabled=True):
        """
        regiter a mesh to polyscope
        :param ps: polyscope instance
        :param name: name of the mesh
        :param v: vertices
        :param f: faces
        :param transparency: transparency of the mesh
        :param edge_width: edge width of the mesh
        :param smooth_shade: smooth shading
        :param c_vertices: vertex scalar values
        :param c_faces: face scalar values
        :param s_faces: face colors
        :param v_vertices: vertex vectors
        """
        ps_mesh = self.ps.register_surface_mesh(name, v, f,
                                        edge_width=edge_width,
                                        transparency=transparency,
                                        color=color,
                                        smooth_shade=smooth_shade,
                                        enabled=enabled)
        if cmap is None:
            cmap = "reds"
        if c_vertices is not None:
            c = (c_vertices - np.min(c_vertices)) / (np.max(c_vertices) - np.min(c_vertices))
            ps_mesh.add_scalar_quantity("vscalar", c,
                                        defined_on="vertices",
                                        enabled=True,
                                        vminmax=(0., 1.),
                                        cmap=cmap)
        if c_faces is not None:
            c = (c_faces - np.min(c_faces)) / (np.max(c_faces) - np.min(c_faces))
            ps_mesh.add_scalar_quantity("fscalar", c,
                                        defined_on="faces",
                                        enabled=True,
                                        vminmax=(0., 1.),
                                        cmap=cmap)
        if s_faces is not None:
            ps_mesh.add_color_quantity("fcolor", s_faces,
                                        defined_on="faces",
                                        enabled=True)
        if v_vertices is not None:
            ps_mesh.add_vector_quantity("vecs", v_vertices,
                                        enabled=True,
                                        # vectortype='ambient',
                                        radius=0.01,
                                        length=0.1,
                                        color=(0.2, 0.5, 0.5))
        return ps_mesh

    def show(self):
        self.ps.show()

gviewer = gviewer()


import numpy as np
from PIL import Image, ImageDraw, ImageFont
from scipy import interpolate, spatial
from .core import to_8b, to_float, to_hom, homogenize, broadcast_batch, map_range
from .structures import get_gizmo_coords

def alpha_compose(images, bg_color=None):
    """
    composes a single or batch of RGBA images into a single or batch of RGB images
    :param image: b x H x W x 4 or H x W x 4
    :param bg_color: 3 or b x 3
    :return: b x H x W x 3 or H x W x 3
    """
    if bg_color is None:
        bg_color = np.array([0., 0., 0.]).astype(np.float32)
    if images.ndim != 3 and images.ndim != 4:
        raise ValueError("image must be 3 or 4 dimensional")
    if images.shape[-1] != 4:
        raise ValueError("image must have 4 channels")
    if images.dtype != np.float32:
        images = to_float(images)
    alpha = images[..., 3:4]
    rgb = images[..., :3]
    return alpha * rgb + (1 - alpha) * bg_color

def draw_text_on_image(images, text_per_image, fill_white=True):
    """
    writes text on images given as np array (b x H x W x 3)
    :param images: (b x H x W x 3) numpy array
    :param text_per_image: b x 1 numpy array of strings
    :param fill_white: if True, text is white, otherwise black
    :return: new (b x H x W x 3) numpy array with text written
    """
    is_float = images.dtype == np.float32
    if is_float:
        images = to_8b(images)
    rgbs = [Image.fromarray(x) for x in images]
    font = ImageFont.truetype("./FreeMono.ttf", 48)
    if fill_white:
        fill = "white"
    else:
        fill = "black"
    for i, rgb in enumerate(rgbs):
        text = text_per_image[i]
        ImageDraw.Draw(rgb).text((0, 0), text, fill=fill, font=font)
    rgbs = np.array([np.asarray(rgb) for rgb in rgbs])
    if is_float:
        rgbs = to_float(rgbs)
    return rgbs

def draw_gizmo_on_image(np_images, w2c, isOpenGL=False, scale=.05):
    """
    adds a gizmo to a batch of np images.
    note: will broadcast np_images and w2c against eachother.
    :param np_images: b x H x W x 3
    :param w2c: b x 3 x 4 w2c transforms (opencv conventions)
    :param isOpenGL: if True, the w2c transforms are assumed to be in OpenGL conventions, else OpenCV conventions
    :param scale: scale of the gizmo
    :return: b x H x W x 3
    """
    new_images = []
    if np_images.ndim != 4:
        raise ValueError("np_images must be b x H x W x 3")
    if w2c.ndim != 3:
        raise ValueError("KRt must be b x 3 x 4")
    np_images, w2c = broadcast_batch(np_images, w2c)
    for i, np_image in enumerate(np_images):
        pil_image = Image.fromarray(to_8b(np_image))
        W, H = pil_image.size
        # W, H = image.shape[1], image.shape[0]
        gizmo_cords = get_gizmo_coords(scale)
        gizmo_hom = to_hom(gizmo_cords)  #  = np.concatenate((gizmo_cords, np.ones_like(gizmo_cords[:, 0:1])), axis=-1)
        verts_clip = (w2c[i] @ gizmo_hom.T).T
        verts_clip = homogenize(verts_clip) # verts_screen_xy = verts_screen[:, :2] / verts_screen[:, 2:3]
        if isOpenGL:
            raise NotImplementedError("OpenGL convention is not supported for now")
            # if not (np.abs(verts_screen_xy[:, 2]) <= 1).all():
            #    raise ValueError("OpenGL convention is not followed")
            # verts_clip = verts_clip[:, :2]
            # verts_screen = np.array([W, H]) * (verts_clip + 1) / 2
        else:
            verts_screen = verts_clip
        desired_loc = np.array([W - 40, H - 40])
        verts_screen += desired_loc - verts_screen[0]
        draw = ImageDraw.Draw(pil_image)
        draw.line((tuple(verts_screen[0]), tuple(verts_screen[1])), fill="red", width = 0)
        draw.line((tuple(verts_screen[0]), tuple(verts_screen[2])), fill="green", width = 0)
        draw.line((tuple(verts_screen[0]), tuple(verts_screen[3])), fill="blue", width = 0)
        new_images.append(np.array(pil_image))
    return (np.array(new_images) / 255.).astype(np.float32)

def merge_figures_with_line(img1, img2, lower_intersection=0.6, angle=np.pi/2, line_width=5, line_color=[255, 255, 255, 255]):
    """
    merges two np images (H x W x 3) with a white line in between
    :param img1: (H x W x 3) numpy array
    :param img2: (H x W x 3) numpy array
    :param lower_intersection: lower intersection of the line with the images
    :param angle: angle of the line
    :param line_width: width of the line
    :param line_color: color of the line
    :return: new (H x W x 3) numpy array with line in between
    """
    if img1.dtype != np.uint8:
        line_color = np.array(line_color) / 255
    else:
        line_color = np.array(line_color, dtype=np.uint8)
    combined = np.ascontiguousarray(np.broadcast_to(line_color, img1.shape))
    # combined = np.ones_like(img1)*255
    y, x, _ = img1.shape
    yy, xx = np.mgrid[:y, :x]
    img1_positions = (xx-lower_intersection*x)*np.tan(angle)-line_width//2>(yy-y)
    img2_positions = (xx-lower_intersection*x)*np.tan(angle)+line_width//2<(yy-y)
    combined[img1_positions] = img1[img1_positions]
    combined[img2_positions] = img2[img2_positions]
    return combined

def generate_voronoi_diagram(height, width, num_cells=1000, bg_color="white", dst=None):
    """
    generate a voronoi diagram HxWx3 with random colored cells
    :param height: height of the image
    :param width: width of the image
    :param num_cells: number of cells
    :param bg_color: background color
    :param dst: if not None, the image is written to this path
    :return: (H x W x 3) numpy array
    """
    nx = np.random.rand(num_cells) * width
    ny = np.random.rand(num_cells) * height
    nxy = np.stack((nx, ny), axis=-1)
    img = Image.new("RGB", (width, height), bg_color)
    vor = spatial.Voronoi(nxy)
    polys = vor.regions
    vertices = vor.vertices
    for poly in polys:
        polygon = vertices[poly]
        if len(poly) > 0 and np.all(np.array(poly) > 0):
            img1 = ImageDraw.Draw(img)
            img1.polygon(list(map(tuple, polygon)), fill=tuple(np.random.randint(0, 255, size=(3,))))
    if dst is not None:
        img.save(str(dst))
    return np.array(img)

def interpolate_single_channel(image: np.ndarray, mask: np.ndarray, method: str = "linear", fill_value: int = 0):
    """
    :param image: (H x W) numpy array
    :param mask: (H x W) boolean numpy array, True indicates missing values
    :param method: interpolation method, one of
        'nearest', 'linear', 'cubic'.
    :param fill_value: which value to use for filling up data outside the
        convex hull of known pixel values.
        Default is 0, Has no effect for 'nearest'.
    :return: the image with missing values interpolated
    """
    h, w = image.shape[:2]
    xx, yy = np.meshgrid(np.arange(w), np.arange(h))

    known_x = xx[~mask]
    known_y = yy[~mask]
    known_v = image[~mask]
    missing_x = xx[mask]
    missing_y = yy[mask]

    interp_values = interpolate.griddata(
        (known_x, known_y), known_v, (missing_x, missing_y),
        method=method, fill_value=fill_value
    )

    interp_image = image.copy()
    interp_image[missing_y, missing_x] = interp_values

    return interp_image

def interpolate_multi_channel(image: np.ndarray, mask: np.ndarray):
    """
    given a multi channel image and a mask, interpolate the values where mask is true (per channel interpolation)
    :param image: (H x W x C) numpy array
    :param mask: (H x W) numpy array
    :return: (H x W x C) numpy array
    """
    if image.ndim != 3:
        raise ValueError("image must have atleast 1 channel")
    interpolated = np.zeros_like(image)
    for channel in range(image.shape[-1]):
        interpolated_channel = interpolate_single_channel(image[:, :, channel], mask)
        interpolated[:, :, channel] = interpolated_channel
    return interpolated

def image_grid(images, rows, cols):
    """
    :param images: list of images
    :param rows: number of rows
    :param cols: number of cols
    :return: grid image
    """
    if images.ndim != 4:
        raise ValueError("images must be a 4D array")
    if len(images) != rows * cols:
        raise ValueError("number of images must be equal to rows * cols")
    tmp = images.reshape(rows, cols, images.shape[1], images.shape[2], -1)
    result = tmp.transpose(0, 2, 1, 3, 4).reshape(rows * images.shape[1], cols * images.shape[2], -1)
    return result

def resize_images_naive(images, H, W, channels_last=True, mode="mean"):
    """
    resize images to output_size, but only if the output size has a common divisor of the input size
    :param images: numpy array of images (N x H x W x C)
    :param H: output height size that has a common divisor with the input height size
    :param W: output width size that has a common divisor with the input width size
    :param channels_last: if True, the images are provided in channels last format (and so will the output)
    :param mode: one of "max", "mean"
    :return: np array of resized images
    """
    if images.ndim != 4:
        raise ValueError("images must be a 4D array")
    if images.shape[1] != images.shape[2]:
        raise ValueError("images must be square")
    if not channels_last:
        raise NotImplementedError("only channels last is supported")
    input_size = np.array(images.shape[1:3])
    output_size = np.array([H, W])
    bin_size = input_size // output_size
    if mode == "max":
        small_images = images.reshape((images.shape[0], output_size[0], bin_size[0], output_size[1], bin_size[1], 3)).max(4).max(2)
    elif mode == "mean":
        small_images = images.reshape((images.shape[0], output_size[0], bin_size[0], output_size[1], bin_size[1], 3)).mean(4).mean(2)
        small_images = small_images.astype(np.uint8)
    else:
        raise ValueError("mode must be one of 'max', 'mean'")
    return small_images

def change_brightness(input_img, brightness=255):
    """
    changes brightness of an image
    :param input_img the numpy image
    :param brightness a number between -255 to 255 (0=no change)
    :return the new image
    """
    brightness = map_range(brightness, 0, 510, -255, 255)

    if brightness != 0:
        if brightness > 0:
            shadow = brightness
            highlight = 255
        else:
            shadow = 0
            highlight = 255 + brightness
        alpha_b = (highlight - shadow)/255
        gamma_b = shadow
    else:
        alpha_b = 1
        gamma_b = 0
    return input_img*alpha_b + gamma_b

def change_contrast(input_img, contrast=127):
    """
    changes brightness of an image
    :param input_img the numpy image
    :param brightness a number between -127 to 127 (0=no change)
    :return the new image
    """
    contrast = map_range(contrast, 0, 254, -127, 127)
    if contrast != 0:
        f = float(131 * (contrast + 127)) / (127 * (131 - contrast))
        alpha_c = f
        gamma_c = 127*(1-f)
    else:
        alpha_c = 1
        gamma_c = 0
    return input_img*alpha_c + gamma_c


import torch
import numpy as np
from pathlib import Path
from .core import to_8b, to_np
from PIL import Image
import json

def write_to_json(data, dst):
    """
    writes data to json file
    :param data: data to write
    :param dst: path to save json file to
    """
    dst = Path(dst)
    dst.parent.mkdir(parents=True, exist_ok=True)
    with open(dst, "w") as f:
        json.dump(data, f, indent=4, sort_keys=True)


def save_animation(images, dst):
    """
    saves a gif animation
    :param images: (b x H x W x C) tensor
    :param dst: path to save animation to
    """
    dst = Path(dst)
    if type(images) == torch.Tensor:
        images = to_np(images)
    if np.isnan(images).any():
        raise ValueError("Images must be finite")
    if images.dtype != np.uint8:
        images = to_8b(images)
    if images.shape[-1] == 1:
        images = [Image.fromarray(image[..., 0], mode="L").convert('P') for image in images]
    else:
        images = [Image.fromarray(image) for image in images]
    dst = Path(dst.parent, dst.stem)
    images[0].save(str(dst)+".gif", save_all=True, append_images=images[1:], optimize=False, duration=100, loop=0)

def save_image(image, dst, force_grayscale: bool = False):
    """
    saves single image as png
    :param image: (H x W x C) tensor
    :param dst: path to save image to (full path to destination, suffix not neccessary but allowed)
    :param force_grayscale: if True, saves image as grayscale
    :param file_name: if provided, saves image with this name
    """
    if image.ndim != 3:
        raise ValueError("Image must be 3 dimensional")
    dst = Path(dst)
    dst.parent.mkdir(parents=True, exist_ok=True)
    save_images(image[None, ...], dst.parent, [dst.name], force_grayscale)

def save_images(images, dst, file_names: list = [], force_grayscale: bool = False, overwrite: bool = True):
    """
    saves images as png
    :param images: (b x H x W x C) tensor
    :param dst: path to save images to
    :param force_grayscale: if True, saves images as grayscale
    :param file_names: if provided, saves images with these names (list of length b)
    """
    if type(images) == torch.Tensor:
        images = to_np(images)
    if np.isnan(images).any():
        raise ValueError("Images must be finite")
    if images.dtype == np.float32 or images.dtype == np.float64:
        images = to_8b(images)
    if images.ndim != 4:
        raise ValueError("Images must be of shape (b x H x W x C)")
    if file_names:
        if images.shape[0] != len(file_names):
            raise ValueError("Number of images and length of file names list must match")
        file_names = [Path(x).stem for x in file_names]  # remove suffix
    for i, image in enumerate(images):
        if force_grayscale or images.shape[-1] == 1:
            if images.shape[-1] == 3:
                image = image.mean(axis=-1, keepdims=True).astype(np.uint8)
            pil_image = Image.fromarray(image[..., 0], mode="L")
        else:
            pil_image = Image.fromarray(image)
        if file_names is not None:
            cur_dst = Path(dst, "{}.png".format(file_names[i]))
            if not overwrite:
                if cur_dst.exists():
                    continue
            pil_image.save(str(cur_dst))
        else:
            cur_dst = Path(dst, "{:05d}.png".format(i))
            if not overwrite:
                if cur_dst.exists():
                    continue
            pil_image.save(str(cur_dst))

def load_image(path, to_float=False, channels_last=True, to_torch=False, device=None, resize_wh=None, as_grayscale=False):
    """
    loads an image from a single file
    :param path: path to file
    :param to_float: if True, converts image to float
    :param return_paths: if True, returns a list of file paths
    :param to_torch: if True, returns a torch tensor
    :param device: device to load tensor to
    :param resize_wh: a tuple (w, h) to resize the image using nearest neighbor interpolation
    :param as_grayscale: if True, loads image as grayscale by averaging over the channels
    :return: (b x H x W x 3) tensor, and optionally a list of file names
    """
    path = Path(path)
    if not path.exists():
        raise FileNotFoundError("Path does not exist")
    if path.is_dir():
        raise FileNotFoundError("Path must be a file")
    elif path.is_file():
        image = load_images(path, to_float=to_float, channels_last=channels_last, return_paths=False, to_torch=to_torch, device=device, resize_wh=resize_wh, as_grayscale=as_grayscale)
        return image[0]

def load_images(source, to_float=False, channels_last=True, return_paths=False, to_torch=False, device=None, resize_wh=None, as_grayscale=False):
    """
    loads images from a list of paths, a folder or a single file
    :param source: path to folder with images / path to image file / list of paths
    :param to_float: if True, converts images to float (and normalizes to [0, 1])
    :param return_paths: if True, returns a list of file paths
    :param to_torch: if True, returns a torch tensor
    :param device: device to load tensor to
    :param resize_wh: a tuple (w, h) to resize all images using nearest neighbor interpolation
    :param as_grayscale: if True, loads images as grayscale by averaging over the channels
    :return: (b x H x W x C) tensor, and optionally a list of file names
    """
    supported_suffixes = [".png", ".jpg", ".jpeg"]
    if type(source) == list or type(source) == tuple or type(source) == np.ndarray:
        images = []
        file_paths = []
        for p in source:
            p = Path(p)
            if not p.exists():
                raise FileNotFoundError("Path does not exist")
            if p.suffix in supported_suffixes:
                im = Image.open(str(p))
                if resize_wh is not None:
                    im = im.resize(resize_wh)
                images.append(np.array(im))
                file_paths.append(p)
        images = np.stack(images, axis=0)
        if as_grayscale and images.ndim == 4:
            images = images.mean(axis=-1).astype(np.float32)
        if not channels_last and images.ndim == 4:
            images = np.moveaxis(images, -1, 1)
        if to_float:
            images = images.astype(np.float32) / 255
        else:
            images = images.astype(np.uint8)
        if to_torch:
            if device is None:
                device = torch.device("cpu")
            images = torch.tensor(images, device=device)
    else:
        path = Path(source)
        if not path.exists():
            raise FileNotFoundError("Path does not exist")
        if path.is_dir():
            images = []
            file_paths = []
            for image in sorted(path.iterdir()):
                if image.suffix in supported_suffixes:
                    im = Image.open(str(image))
                    if resize_wh is not None:
                        im = im.resize(resize_wh)
                    images.append(np.array(im))
                    file_paths.append(image)
            images = np.stack(images, axis=0)
            if as_grayscale and images.ndim == 4:
                images = images.mean(axis=-1).astype(np.float32)
            if not channels_last and images.ndim == 4:
                images = np.moveaxis(images, -1, 1)
            if to_float:
                images = images.astype(np.float32) / 255
            else:
                images = images.astype(np.uint8)
            if to_torch:
                if device is None:
                    device = torch.device("cpu")
                images = torch.tensor(images, device=device)
        elif path.is_file():
            im = Image.open(str(path))
            if resize_wh is not None:
                im = im.resize(resize_wh)
            images = np.array(im)
            if as_grayscale and images.ndim == 3:
                images = images.mean(axis=-1).astype(np.float32)
            if not channels_last and images.ndim == 3:
                images = np.moveaxis(images, -1, 0)
            if to_float:
                images = images.astype(np.float32) / 255
            else:
                images = images.astype(np.uint8)
            file_paths = [path]
            if to_torch:
                if device is None:
                    device = torch.device("cpu")
                images = torch.tensor(images, device=device)
            images = images[None, ...]
    if return_paths:
        return images, file_paths
    else:
        return images

def load_mesh(path: Path, to_torch=False, device=None, verbose=True):
    """
    loads a mesh from a file
    :param path: path to mesh file
    :param to_torch: if True, returns a torch tensor
    :param device: device to load tensor to
    :return: (V x 3) tensor of vertices, (F x 3) tensor of faces, and optionally (V x 3) tensor of normals
    """
    path = Path(path)
    if path.suffix != ".obj":
        raise ValueError("Only .obj are supported")
    return load_obj(path, to_torch=to_torch, device=device, verbose=verbose)

def load_obj(path: Path, to_torch=False, device=None, verbose=True):
    """
    :param path: path to obj file
    :param to_torch: if True, returns a torch tensor
    :param device: device to load tensor to
    :return: (V x 3) tensor, (F x 3) tensor, and optionally (V x 3) tensor
    """
    path = Path(path)
    if not path.exists():
        raise ValueError("Path does not exist")
    if not path.is_file():
        raise ValueError("Path must be a file")
    if path.suffix != ".obj":
        raise ValueError("Only .obj are supported")
    v, f = parse_obj(path, verbose=verbose)
    if to_torch and device is not None:
        v = torch.tensor(v, dtype=torch.float, device=device)
        f = torch.tensor(f,dtype=torch.long, device=device)
        # n = torch.tensor(n, dtype=torch.float, device=device)
    return v, f

def parse_obj(path: Path, verbose=True):
    """
    A simple obj parser
    currently supports vertex and triangular face elements only
    """
    path = Path(path)
    vertexList = []
    vertexTextureList = []
    vertexNormalList = []
    colorList = []
    faceList = []
    finite_flag = False
    with open(path, 'r') as objFile:
        for line in objFile:
            line = line.split("#")[0]  # remove comments
            split = line.split()
            if not len(split):  # skip empty lines
                continue
            if split[0] == "v":
                if 3 <= len(split[1:]) <= 4:  # x y z [w]
                    float_vertex = np.array([np.float64(x) for x in split[1:]])
                    if not np.isfinite(float_vertex).all():
                        finite_flag=True
                    vertexList.append(float_vertex)
                elif len(split[1:]) == 6:
                    float_vertex = np.array([np.float64(x) for x in split[1:4]])
                    float_color = np.array([np.float64(x) for x in split[4:]])
                    vertexList.append(float_vertex)
                    colorList.append(float_color)
                else:
                    raise ValueError("vertex {} has {} entries, but only 3 or 6 are supported".format(len(vertexList), len(split[1:])))
            elif split[0] == "f":
                if len(split[1:]) != 3:
                    raise ValueError("only triangular faces are supported")
                else:
                    face = [x.split("/")[0] for x in split[1:]]
                    int_face = np.array([int(x)-1 for x in face])
                    if np.any(int_face < 0):
                        raise ValueError("negative face indices are not supported")
                    faceList.append(int_face)
            elif split[0] == "vn":
                if len(split[1:]) != 3:  # xn yn zn
                    raise ValueError("vertex normal {} has {} entries, but only 3 are supported".format(len(vertexNormalList), len(split[1:])))
                else:
                    float_vertex_normal = np.array([np.float64(x) for x in split[1:]])
                    vertexNormalList.append(float_vertex_normal)
            elif split[0] == "vt":
                if 2 <= len(split[1:]) <= 3:  # u [v, w]
                    float_vertex_tex = np.array([np.float64(x) for x in split[1:]])
                    if (float_vertex_tex < 0).any():
                        raise ValueError("negative texture coordinates are not supported")
                    if (float_vertex_tex > 1).any():
                        raise ValueError("texture coordinates must be between 0 and 1")
                    vertexTextureList.append(float_vertex_tex)
                else:
                    raise ValueError("vertex normal {} has {} entries, but only 3 are supported".format(len(vertexNormalList), len(split[1:])))
            elif split[0] == "l":  # ignore polyline elements
                continue
            elif split[0] == "vp":  # ignore parameter space vertices
                continue
    v = np.stack(vertexList)
    f = np.stack(faceList)
    if finite_flag and verbose:
        print("Warning: some vertices in file were not finite")
    # vn = np.stack(vertexNormalList)
    # vt = np.stack(vertexTextureList)
    return v, f #, vn, vt

def save_obj(path: Path, vertices, faces):
    """"
    :param path: path to save obj file to
    :param vertices: (n x 3) tensor of vertices
    :param faces: (m x 3) tensor of vertex indices
    """
    path = Path(path)
    if path.suffix != ".obj":
        raise ValueError("Path must have suffix .obj")
    if not path.parent.exists():
        path.parent.mkdir(parents=True, exist_ok=True)
    if type(vertices) == torch.Tensor:
        vertices = to_np(vertices)
    if type(faces) == torch.Tensor:
        faces = to_np(faces)
    if (faces < 0).any():
        raise ValueError("Faces must be positive")
    if np.isnan(vertices).any():
        raise ValueError("Vertices must be finite")
    if np.isnan(faces).any():
        raise ValueError("Faces must be finite")
    if vertices.dtype != np.float32 and vertices.dtype != np.float64:
        raise ValueError("Vertices must be of type float32 / float64")
    if faces.dtype != np.int32 and faces.dtype != np.int64:
        raise ValueError("Faces must be of type int32 / int64")
    with open(str(path), "w") as file:
        for v in vertices:
            file.write("v {} {} {}\n".format(v[0], v[1], v[2]))  # write vertices
        for f in faces:
            file.write("f {} {} {}\n".format(f[0] + 1, f[1] + 1, f[2] + 1))  # obj indices start at 1

def save_ply(path: Path, vertices, faces):
    raise NotImplementedError

def save_mesh(path, vertices, faces):
    """
    saves a mesh to a file
    :param path: path to save mesh to
    :param vertices: (n x 3) tensor of vertices
    :param faces: (m x 3) tensor of vertex indices
    """
    path = Path(path)
    if path.suffix not in [".obj"]:
        raise ValueError("Only .obj is supported")
    else:
        if path.suffix == ".obj":
            save_obj(path, vertices, faces)
        # elif path.suffix == ".ply":
        #     save_ply(path, vertices, faces)

def save_pointcloud(path: Path, vertices):
    path = Path(path)
    if path.suffix != ".ply":
        raise ValueError("Only .ply are supported")
    else:
        save_ply(path, vertices, None)

def save_meshes(path, vertices, faces, file_names: list = []):
    """
    saves a list of meshes to a folder
    :param path: path to save meshes to
    :param vertices: (b x V x 3) tensor
    :param faces: (b x F x 3) tensor
    :param file_names: list of file names (of length b)
    """
    if vertices.ndim != 3 or faces.ndim != 3:
        raise ValueError("Vertices and faces must be 3D tensors")
    if vertices.shape[0] != faces.shape[0]:
        raise ValueError("Vertices and faces must have the same batch size")
    if file_names:
        if len(file_names) != vertices.shape[0]:
            raise ValueError("Number of file names must match batch size")
    path = Path(path)
    for i, (v, f) in enumerate(zip(vertices, faces)):
        if file_names:
            save_mesh(path / "{}.obj".format(file_names[i]), v, f)
        else:
            save_mesh(path / "{:05d}.obj".format(i), v, f)